from rdopkg import helpers
from rdopkg import exception
import pytest


@pytest.mark.skip('Temporarily diabled during migration to sf.')
def test_editor_editor_run_failed(monkeypatch):

    def mock_run(*args, **kwargs):
        raise exception.CommandNotFound('Foo')
    monkeypatch.setattr(helpers, 'run', mock_run)

    def mock_env(var):
        return 'foo_editor'
    monkeypatch.setattr(helpers.os.environ, 'get', mock_env)

    with pytest.raises(exception.CommandNotFound) as excinfo:
        helpers.edit('/tmp/foo')
    assert excinfo.match(r'.*\(foo_editor\).* Please set \$EDITOR .*')


@pytest.mark.skip('Temporarily diabled during migration to sf.')
def test_editor_no_env(monkeypatch):

    def mock_run(*args, **kwargs):
        raise exception.CommandNotFound('Foo')
    monkeypatch.setattr(helpers, 'run', mock_run)

    def mock_env(var):
        return ''
    monkeypatch.setattr(helpers.os.environ, 'get', mock_env)

    with pytest.raises(exception.CommandNotFound) as excinfo:
        helpers.edit('/tmp/foo')
    assert excinfo.match(r'.*\(vim\).* Please set \$EDITOR .*')



# -*- coding: UTF-8 -*-
import os
import shutil


def clean_tempdir(context, scenario):
    """
    Clean up temporary test dirs for passed tests.

    Leave failed test dirs for manual inspection.

    """
    tempdir = getattr(context, 'tempdir', None)
    if tempdir and scenario.status == 'passed':
        shutil.rmtree(tempdir)
        del(context.tempdir)


def restore_cwd(context):
    """
    Return to original working dir or behave gets confused :-/
    """
    old_cwd = getattr(context, 'old_cwd', None)
    if old_cwd:
        os.chdir(old_cwd)


def after_scenario(context, scenario):
    clean_tempdir(context, scenario)
    restore_cwd(context)



"""
This module implement generic support of Koji through Koji ugly API.
For now, it's only used for CBS.
"""
from __future__ import print_function
from distutils.spawn import find_executable
import os
import subprocess

from rdopkg.exception import ModuleNotAvailable, RpmModuleNotAvailable
from rdopkg.utils.specfile import spec_fn, Spec
from rdopkg.utils.log import log
from rdopkg import exception, guess

from bunch import bunchify

KOJI_AVAILABLE = False
try:
    import koji
    KOJI_AVAILABLE = True
except ImportError:
    pass
RPM_AVAILABLE = False
try:
    import rpm
    RPM_AVAILABLE = True
except ImportError:
    pass


def setup_kojiclient(profile):
    """Setup koji client session
    """
    opts = koji.read_config(profile)
    for k, v in opts.iteritems():
        opts[k] = os.path.expanduser(v) if type(v) is str else v
    kojiclient = koji.ClientSession(opts['server'], opts=opts)
    kojiclient.ssl_login(opts['cert'], None, opts['serverca'])
    return kojiclient


def new_build(profile='cbs', scratch=True):
    # Very ugly: some utilities are only available in koji CLI
    # and not in the koji module
    if not KOJI_AVAILABLE:
        raise ModuleNotAvailable(module='koji')
    if not RPM_AVAILABLE:
        raise RpmModuleNotAvailable()
    import imp
    kojibin = find_executable('koji')
    kojicli = imp.load_source('kojicli', kojibin)
    # Koji 1.13 moves client internal API into another path
    try:
        from koji_cli.lib import _unique_path, _progress_callback, watch_tasks
    except ImportError:
        from kojicli import _unique_path, _progress_callback, watch_tasks

    kojiclient = setup_kojiclient(profile)
    # Note: required to make watch_tasks work
    kojicli.options = opts = bunchify(kojiclient.opts)
    build_target = guess_build()
    if not build_target:
        log.warn("failed to identify build tag from branch")
        return

    retrieve_sources()
    srpm = create_srpm()
    if not srpm:
        log.warn('No srpm available')
        return
    serverdir = _unique_path('cli-build')
    kojiclient.uploadWrapper(srpm, serverdir, callback=_progress_callback)
    source = "%s/%s" % (serverdir, os.path.basename(srpm))
    task_id = kojiclient.build(source, build_target, {'scratch': scratch})
    print("Created task:", task_id)

    print("Task info: {}/taskinfo?taskID={}".format(opts['weburl'], task_id))
    kojiclient.logout()
    watch_tasks(kojiclient, [task_id])


def guess_build():
    release = guess.osrelease()
    builds = guess.builds(release)
    if builds:
        build_tag = builds[0][1].replace('cbs/', '')
        return build_tag
    return None


def retrieve_sources():
    """Retrieve sources using spectool
    """
    spectool = find_executable('spectool')
    if not spectool:
        log.warn('spectool is not installed')
        return
    try:
        specfile = spec_fn()
    except Exception:
        return

    cmd = [spectool, "-g", specfile]
    output = subprocess.check_output(' '.join(cmd), shell=True)
    log.warn(output)


def create_srpm(dist='el7'):
    """Create an srpm
    Requires that sources are available in local directory

    dist: set package dist tag (default: el7)
    """
    if not RPM_AVAILABLE:
        raise RpmModuleNotAvailable()
    path = os.getcwd()
    try:
        specfile = spec_fn()
        spec = Spec(specfile)
    except Exception:
        return

    rpmdefines = ["--define 'dist .{}'".format(dist),
                  "--define '_sourcedir {}'".format(path),
                  "--define '_srcrpmdir {}'".format(path)]
    rpm.addMacro('_sourcedir', '.{}'.format(dist))
    # FIXME: needs to be fixed in Spec
    rpm.addMacro('dist', '.{}'.format(dist))
    module_name = spec.get_tag('Name', True)
    version = spec.get_tag('Version', True)
    release = spec.get_tag('Release', True)
    srpm = os.path.join(path,
                        "{}-{}-{}.src.rpm".format(module_name,
                                                  version,
                                                  release))

    # See if we need to build the srpm
    if os.path.exists(srpm):
        log.warn('Srpm found, rewriting it.')

    cmd = ['rpmbuild']
    cmd.extend(rpmdefines)

    cmd.extend(['--nodeps', '-bs', specfile])
    output = subprocess.check_output(' '.join(cmd), shell=True)
    log.warn(output)
    srpm = output.split()[1]
    return srpm



from rdopkg.exception import CommandFailed
from rdopkg.exception import RdopkgException


def test_exception_default():
    e = RdopkgException()
    assert str(e) == 'An unknown error occurred'
    assert e.exit_code == 1


def test_exception_msg_provided():
    e = RdopkgException(msg='foo bar')
    assert str(e) == 'foo bar'


def test_CommandFailed_bad_kwargs():
    e = CommandFailed(foo='bar')
    assert str(e) == e.msg_fmt


def test_CommandFailed_good_kwargs():
    e = CommandFailed(cmd='bar')
    assert str(e) == e.msg_fmt % {'cmd': 'bar'}


def test_exception_exit_code():
    for code in [-128, -1, 0, 1, 23, 1337]:
        e = RdopkgException(exit_code=code)
        assert e.exit_code == code



from __future__ import print_function
from distroinfo.info import DistroInfo
import distroinfo.query
import sys

from rdopkg.actionmods import rdoinfo
from rdopkg import exception
from rdopkg import helpers
from rdopkg.utils import git
from rdopkg.utils import log


def info(pkgs=None, local_info=None, info_file=None,
         apply_tag=None, force_fetch=False):
    if not info_file:
        info_file = rdoinfo.info_file()
    if local_info:
        di = DistroInfo(info_file,
                        local_info=local_info)
    else:
        di = rdoinfo.get_distroinfo()
        if force_fetch:
            di.fetcher.cache_ttl = 0
    _info = di.get_info(apply_tag=apply_tag)
    if pkgs:
        filters = {}
        for pf in pkgs:
            attr, sep, rex = pf.partition(':')
            if not sep:
                # filter by name by default
                attr = 'name'
                rex = pf
            filters[attr] = rex
        rdoinfo.print_pkgs(_info, filters)
    else:
        rdoinfo.print_summary(_info)
        print('')
        print("Supply regex filter(s) to list package details, e.g.:\n{t.cmd}"
              "    rdopkg info nova\n"
              "    rdopkg info conf:client maintainers:jruzicka\n"
              "    rdopkg info '.*'"
              "{t.normal}".format(t=log.term))


def findpkg(query, strict=False, local_info=None, info_file=None,
            force_fetch=False):
    if not info_file:
        info_file = rdoinfo.info_file()
    if local_info:
        di = DistroInfo(info_file, local_info=local_info)
    else:
        di = rdoinfo.get_distroinfo()
        if force_fetch:
            di.fetcher.cache_ttl = 0
    _info = di.get_info()
    pkg = distroinfo.query.find_package(_info, query, strict=strict)
    if not pkg:
        raise exception.InvalidPackage(
            msg="No package found in rdoinfo for query: %s" % query)
    rdoinfo.print_pkg(pkg)


def info_tags_diff(local_info, info_file=None, buildsys_tags=False):
    if not info_file:
        info_file = rdoinfo.info_file()
    di = DistroInfo(info_file, local_info=local_info)
    if buildsys_tags:
        tagsname = 'buildsys-tags'
    else:
        tagsname = 'tags'
    info2 = di.get_info()
    with helpers.cdir(di.fetcher.source):
        with git.git_revision('HEAD~'):
            info1 = di.get_info()
    tdiff = distroinfo.query.tags_diff(info1, info2, tagsname=tagsname)
    if not tdiff:
        sys.stderr.write("No tag changes detected.\n")
    else:
        for pkg, changes in tdiff:
            print("%s %s" % (pkg, changes))



from rdopkg.action import Action, Arg


_SHARED_HELP = {
    'release_bump_index':
        ("specify which Release part to bump:\n"
         "-R MAJOR/-R MINOR/-R PATCH\n"
         "   to bump selected in MAJOR.MINOR.PATCH\n"
         "-R 1/-R 2/-R 3/-R 4/-R N\n"
         "   to bump 1./2./3./4./N. part (numeric index)\n"
         "-R last-numeric\n"
         "   to bump last numeric only Release part (default)"),
}


ACTIONS = [
    Action('clone',
           help="clone an RDO package distgit and setup remotes",
           required_args=[
               Arg('package', positional=True, metavar='PACKAGE',
                   help="RDO package to clone (see `rdopkg info`)"),
           ],
           optional_args=[
               Arg('use_master_distgit', shortcut='-m', action='store_true',
                   help="clone 'master-distgit'"),
               Arg('gerrit_remotes', shortcut='-g', action='store_true',
                   help="create branches "
                        "'gerrit-origin' and 'gerrit-patches'"),
               Arg('review_user', shortcut='-u', metavar='USER',
                   help="gerrit username for reviews"),
               Arg('distro', shortcut='-d', metavar='DISTRO',
                   help="distroinfo configuration"),
           ]),
    Action('pkgenv', help="show detected package environment",
           steps=[
               Action('get_package_env'),
               Action('show_package_env'),
           ]),
    Action('patchlog', help="show patches branch log",
           steps=[
               Action('get_package_env'),
               Action('show_patch_log'),
           ]),
    Action('get-patches',
           help="setup local patches branch and switch to it",
           optional_args=[
               Arg('patches_branch', shortcut='-p', metavar='REMOTE/BRANCH',
                   help="remote git branch containing patches"),
               Arg('local_patches_branch', shortcut='-P',
                   metavar='LOCAL_BRANCH',
                   help="local git branch containing patches"),
               Arg('gerrit_patches_chain', shortcut='-g',
                   metavar='REVIEW_NUMBER',
                   help="top gerrit review id of the patch chain"),
               Arg('force', shortcut='-f', action='store_true',
                   help="use patch even if it was not validated in CI"),
           ],
           steps=[
               Action('get_package_env'),
               Action('ensure_patches_branch'),
               Action('get_patches_branch'),
               Action('checkout_patches_branch'),
           ]),
    Action('fix', continuable=True,
           help="change .spec file without introducing new patches",
           optional_args=[
               Arg('release_bump_index', shortcut='-R', metavar='INDEX',
                   help=_SHARED_HELP['release_bump_index']),
           ],
           steps=[
               Action('get_package_env'),
               Action('update_spec'),
               Action('edit_spec'),
               Action('commit_distgit_update'),
               Action('final_spec_diff'),
           ]),
    Action('patch',
           help="introduce new patches to the package",
           optional_args=[
               Arg('amend', shortcut='-a', action='store_true',
                   help="amend previous commit"),
               Arg('patches_branch', shortcut='-p', metavar='REMOTE/BRANCH',
                   help="remote git branch containing patches"),
               Arg('local_patches_branch', shortcut='-P',
                   metavar='LOCAL_BRANCH',
                   help="local git branch containing patches"),
               Arg('local_patches', shortcut='-l', action='store_true',
                   help="don't reset local patches branch, use it as is"),
               Arg('gerrit_patches_chain', shortcut='-g',
                   metavar='REVIEW_NUMBER',
                   help="top gerrit review id of the patch chain"),
               Arg('force', shortcut='-f', action='store_true',
                   help="use patch even if it was not validated in CI"),
               Arg('release_bump_index', shortcut='-R', metavar='INDEX',
                   help=_SHARED_HELP['release_bump_index']),
               Arg('no_bump', shortcut='-B', action='store_true',
                   help="don't bump release and generate changelog "
                        "(update patches only)"),
               Arg('changelog', shortcut='-C',
                   choices=['detect', 'count', 'plain'],
                   help="how to generate changelog from patches"),
               Arg('commit_header_file', shortcut='-H', metavar='FILE',
                   help="start commit message with FILE contents, "
                        "- for stdin"),
               Arg('changelog_user', shortcut='-u', default=None,
                   help="User to be used in new changelog entry"),
               Arg('changelog_email', shortcut='-e', default=None,
                   help="email address to be used in new changelog entry"),
           ],
           steps=[
               Action('get_package_env'),
               Action('ensure_patches_base_ref'),
               Action('ensure_patches_branch'),
               Action('get_patches_branch'),
               Action('check_new_patches'),
               Action('update_patches'),
               Action('ensure_patches_changed'),
               Action('update_spec'),
               Action('commit_distgit_update'),
               Action('final_spec_diff'),
           ]),
    Action('update_patches',
           alias='patch',
           help='Alias for `rdopkg patch --local-patches --no-bump`',
           description=(
               "A legacy backward compatibility"
               "with ancient update-patches.sh script."
           ),
           const_args={
               'local_patches': True,
               'no_bump': True,
           }),
    Action('new_version', continuable=True,
           help="update package to new upstream version",
           optional_args=[
               Arg('new_version', positional=True, nargs='?',
                   help="version to update to"),
               Arg('patches_branch', shortcut='-p', metavar='REMOTE/BRANCH',
                   help="remote git branch containing patches"),
               Arg('local_patches_branch', shortcut='-P',
                   metavar='LOCAL_BRANCH',
                   help="local git branch containing patches"),
               Arg('local_patches', shortcut='-l', action='store_true',
                   help="don't reset local patches branch, use it as is"),
               Arg('bump_only', shortcut='-b', action='store_true',
                   help="only bump .spec to new version a la rpmdev-bumpspec"),
               Arg('no_diff', shortcut='-d', action='store_true',
                   help="don't show git/requirements diff"),
               Arg('new_sources', shortcut='-N', action='store_true',
                   help=("run `fedpkg new-sources`"
                         " (default: depends on branch name)")),
               Arg('no_new_sources', shortcut='-n', action='store_true',
                   help=("don't run `fedpkg new-sources`"
                         " (default: depends on branch name)")),
               Arg('unattended', shortcut='-U', action='store_true',
                   help="don't ask any questions (NOT RECOMMENDED)"),
               Arg('no_push_patches', shortcut='-t', action='store_true',
                   help="don't push patches branch"),
               Arg('bug', shortcut='-B', metavar='BUG(S)',
                   help="reference BUG(S) in changelog. (example:"
                        " --bug rhbz#1234,rhbz#5678)"),
               Arg('commit_header_file', shortcut='-H', metavar='FILE',
                   help="start commit message with FILE contents, "
                        "- for stdin"),
               Arg('changelog_user', shortcut='-u', default=None,
                   help="User to be used in new changelog entry"),
               Arg('changelog_email', shortcut='-e', default=None,
                   help="email address to be used in new changelog entry"),
           ],
           steps=[
               Action('get_package_env'),
               Action('ensure_patches_base_ref'),
               Action('new_version_setup'),
               Action('diff'),
               Action('prep_new_patches_branch'),
               Action('get_patches_branch'),
               Action('rebase_patches_branch'),
               Action('update_patches'),
               Action('update_spec'),
               Action('get_source'),
               Action('new_sources'),
               Action('commit_distgit_update'),
               Action('final_spec_diff'),
               Action('review_patches_branch')
           ]),
    Action('amend',
           help="amend last commit and recreate commit message",
           optional_args=[
               Arg('commit_header_file', shortcut='-H', metavar='FILE',
                   help="start commit message with FILE contents, "
                        "- for stdin"),
           ]),
    Action('squash',
           help="squash HEAD into HEAD~ using HEAD~ commit message"),
    Action('get_source', help="fetch source archive",
           steps=[
               Action('get_source', const_args={'new_sources': True})
           ]),
    Action('tag_patches',
           help='tag the -patches branch in Git with the current NVR',
           optional_args=[
               Arg('force', shortcut='-f', action='store_true',
                   help='replace an existing tag with this name'),
               Arg('push', shortcut='-p', action='store_true',
                   help='push this new tag to the patches remote'),
           ],
           steps=[
               Action('get_package_env'),
               Action('ensure_patches_branch'),
               Action('tag_patches_branch'),
           ]),
]



from rdopkg import utils


def test_url_tidying():
    # Do nothing when nothing is provided
    assert utils.tidy_ssh_user() is None
    # Do nothing when url is not provided
    assert utils.tidy_ssh_user(url=None, user='kyle') is None
    # Do nothing when url doesn't start with ssh
    uri = 'http://mygitrepo/myproject.git'
    assert utils.tidy_ssh_user(uri) == uri
    assert utils.tidy_ssh_user(uri, user='stan') == uri
    # Do nothing if url starts with ssh, has no user, and no user provided
    uri = 'ssh://%smygitrepo/myproject.git'
    assert utils.tidy_ssh_user(uri % '') == uri % ''
    # Add missing user if provided
    assert utils.tidy_ssh_user(uri % '', user='kenny') == uri % 'kenny@'
    # keep default user if no user provided
    assert utils.tidy_ssh_user(uri % 'cartman@') == uri % 'cartman@'
    # change default user if user provided
    assert utils.tidy_ssh_user(uri % 'timmy@', user='dude') == uri % 'dude@'


def test_gerrit_query(monkeypatch):
    mock_result = """{"project":"ironic","branch":"liberty-patches","topic":"p1","id":"Ie0db0345c9d5f9ad3d2ec4880e084cd38d06a6dc","number":"368","subject":"Set default DB location","owner":{"name":"Fabien Boucher","email":"fabien.dot.boucher@gmail.com","username":"morucci"},"url":"http://rpmfactory.beta.rdoproject.org/r/368","commitMessage":"Set default DB location","createdOn":1453933391,"lastUpdated":1456332266,"open":true,"status":"NEW"}
{"type":"stats","rowCount":1,"runTimeMilliseconds":4,"moreChanges":true}"""  # noqa

    def mock_run(*args, **kwargs):
        return mock_result
    monkeypatch.setattr(utils.cmd, 'run', mock_run)
    g = utils.cmd.GerritQuery('gerrithost', '29418')
    r = g('project:ironic')
    assert isinstance(r, dict) is True
    assert r['id'] == "Ie0db0345c9d5f9ad3d2ec4880e084cd38d06a6dc"
    mock_result = """{"type":"stats","rowCount":0,"runTimeMilliseconds":4,"moreChanges":false}"""  # noqa
    r = g('project:ironic')
    assert r is None



import sys

from . import __version__
from rdopkg.action import ActionManager
from rdopkg.core import ActionRunner
from rdopkg import shell
from rdopkg import actions


def rdopkg_runner():
    """
    default rdopkg action runner including rdopkg action modules
    """
    aman = ActionManager()
    # assume all actions.* modules are action modules
    aman.add_actions_modules(actions)
    aman.fill_aliases()
    # additional rdopkg action module logic should go here
    return ActionRunner(action_manager=aman)


def rdopkg(*cargs):
    """
    rdopkg CLI interface

    Execute rdopkg action with specified arguments and return
    shell friendly exit code.

    This is the default high level way to interact with rdopkg.

        py> rdopkg('new-version', '1.2.3')

    is equivalent to

        $> rdopkg new-version 1.2.3
    """
    runner = rdopkg_runner()
    return shell.run(runner,
                     cargs=cargs,
                     prog='rdopkg',
                     version=__version__)


def main():
    """
    rdopkg console_scripts entry point
    """
    cargs = sys.argv[1:]
    sys.exit(rdopkg(*cargs))


if __name__ == '__main__':
    main()



"""
Package build related actions.
"""


from rdopkg.actions.distgit.actions import FEDPKG
from rdopkg import exception
from rdopkg import guess
from rdopkg.actionmods import cbsbuild
from rdopkg.actionmods import kojibuild
from rdopkg.utils import log
from rdopkg.utils.git import git
from rdopkg.utils.cmd import run
from rdopkg import helpers


def fedpkg_mockbuild(fedpkg=FEDPKG):
    cmd = list(fedpkg) + ['mockbuild']
    run(*cmd, direct=True)


def koji_build():
    if git.branch_needs_push():
        helpers.confirm("It seems local distgit branch needs push. Push "
                        "now?")
        git('push')
    kojibuild.new_build()


def cbs_build(scratch=False):
    if git.branch_needs_push() and not scratch:
        helpers.confirm("It seems local distgit branch needs push. Push "
                        "now?")
        git('push')
    cbsbuild.new_build(profile='cbs', scratch=scratch)



import difflib


def exdiff(txt1, txt2, header=None):
    d = difflib.ndiff(txt1.splitlines(1), txt2.splitlines(2))
    s = ''.join(list(d))
    if header:
        s = "%s\n%s" % (header, s)
    # python 2 exceptions fail with unicode chars (?!)
    s = s.encode('ascii', 'replace')
    return s



from behave import *
import os
import re
import tempfile

from rdopkg.utils.distgitmagic import run
from rdopkg.const import STATE_FILE_FN


@when('I run rdopkg {args}')
def step_impl(context, args):
    # proper argument parsing might be needed
    cmd = ['rdopkg'] + args.split(' ')
    context.command_output = run(*cmd, fatal=False, log_fail=False)


@given('a temporary directory')
def step_impl(context):
    context.tempdir = tempfile.mkdtemp(prefix='rdopkg-test-')
    context.old_cwd = os.getcwd()
    os.chdir(context.tempdir)


@then("command output contains '{rex}'")
def step_impl(context, rex):
    assert re.search(rex, context.command_output), \
        "Did not find [{0}] in command output [{1}]".format(
            rex, context.command_output.encode('ascii', 'replace'))


@then("command return code is {n:n}")
def step_impl(context, n):
    rc = context.command_output.return_code
    assert n == rc, \
        "Command return code is %s but should be %s" % (rc, n)


@then(u'rdopkg state file is present')
def step_check_for_rdopkg_state_file_not_present(context):
    assert os.path.exists(os.path.join(context.distgitdir, STATE_FILE_FN))


@then(u'rdopkg state file is not present')
def step_check_for_rdopkg_state_file_not_present(context):
    assert not os.path.exists(os.path.join(context.distgitdir, STATE_FILE_FN))



from rdopkg.action import Action


ACTIONS = [
    Action('status', help="show status of previous action"),
    Action('conf', help="show rdopkg configuration"),
    Action('actions',
           help="list action functions and their availability"),
    Action('autocomplete',
           help="get TAB completion for rdopkg!"),
    Action('doctor',
           help="activate rdopkg psychoanalyst mode"),
]



"""
Because fedpkg/pyrpkg sucks horribly, this entire file full of hacks is
needed to interface with it. Disgusting.
"""
from __future__ import print_function
from six.moves import configparser
import logging
import sys
import os

from rdopkg import exception
from rdopkg.utils.log import log


KOJI_AVAILABLE = False
try:
    import koji  # NOQA
    KOJI_AVAILABLE = True
except ImportError:
    pass

FEDPKG_AVAILABLE = False
try:
    import fedpkg
    FEDPKG_AVAILABLE = True
except ImportError:
    pass


KOJI_HUB_URL = 'http://koji.fedoraproject.org/kojihub'
KOJI_BASE_URL = 'http://kojipkgs.fedoraproject.org/work/'


def modules_check():
    if not KOJI_AVAILABLE:
        raise exception.ModuleNotAvailable(module='koji')
    if not FEDPKG_AVAILABLE:
        raise exception.ModuleNotAvailable(module='fedpkg')


def setup_fedpkg_logger():
    flog = logging.getLogger('rpkg')
    fmtr = logging.Formatter('%(message)s')
    for h in list(flog.handlers):
        flog.removeHandler(h)
    stdouth = logging.StreamHandler(sys.stdout)
    stdouth.setFormatter(fmtr)
    flog.addHandler(stdouth)
    flog.setLevel(logging.INFO)
    return flog


def get_fedpkg_config():
    fedpkg_conf = '/etc/rpkg/fedpkg.conf'
    config = configparser.SafeConfigParser()
    config.read(fedpkg_conf)
    return config


def get_fedpkg_commands():
    config = get_fedpkg_config()
    items = dict(config.items('fedpkg', raw=True))
    # I don't even...
    cmd = fedpkg.Commands(os.getcwd(),
                          items['lookaside'],
                          items['lookasidehash'],
                          items['lookaside_cgi'],
                          items['gitbaseurl'],
                          items['anongiturl'],
                          items['branchre'],
                          items['kojiconfig'],
                          items['build_client'],
                          )
    return cmd


class FedpkgArgsStub(object):
    """
    fedpkg sucks, this is one of many results
    """

    def __init__(self):
        self.q = False


def get_fedpkg_cli(config=None, flog=None):
    if not config:
        config = get_fedpkg_config()
    if not flog:
        flog = logging.getLogger('rpkg')
    cli = fedpkg.cli.cliClient(config)
    cli.log = flog
    cli.args = cli.args = FedpkgArgsStub()
    return cli


def new_build(watch=True):
    modules_check()
    flog = setup_fedpkg_logger()  # NOQA
    fcmd = get_fedpkg_commands()
    task_id = fcmd.build()

    if watch:
        print('')
        try:
            cli = get_fedpkg_cli()
            # TODO: might be good to push this return data back up
            #       or check the status of the return
            r = cli._watch_koji_tasks(fcmd.kojisession, [task_id])
        except configparser.NoSectionError:
            # <C-C> causes this for some reason
            print('')
        except Exception as ex:
            log.warn(str(type(ex).__name__))
            log.warn("Failed to get build status: %s" % ex)

    return fcmd.nvr



from __future__ import print_function
import inspect
import pkgutil
import six

from rdopkg import exception
from rdopkg.utils import log

if six.PY2:
    inspect_getargspec = inspect.getargspec
else:
    inspect_getargspec = inspect.getfullargspec


class Arg(object):
    def __init__(self, name, **kwargs):
        self.name = name
        self.shortcut = kwargs.pop('shortcut', None)
        self.positional = kwargs.pop('positional', False)
        self.kwargs = kwargs

    def cmdarg_names(self, name2argfun):
        names = []
        if self.shortcut:
            names.append(self.shortcut)
        names.append(name2argfun(self.name))
        return names


class Action(object):
    def __init__(self, name, steps=None, alias=None, continuable=False,
                 required_args=None, optional_args=None, const_args=None,
                 help=None, description=None, module=None, action_fun=None):
        if not const_args:
            const_args = []
        if not required_args:
            required_args = []
        if not optional_args:
            optional_args = []
        self.name = name
        self.module = module
        self.steps = steps
        self.alias = alias
        self.continuable = continuable
        self.action_fun = action_fun
        self.const_args = const_args
        self.required_args = required_args
        self.optional_args = optional_args
        self.help = help
        self.description = description

    def __repr__(self):
        return "ACTION: %s" % self.name


class ActionModule(object):
    def __init__(self, module, name=None):
        if not name:
            name = module.__name__
        self.name = name
        self.module = module
        self.actions = module.ACTIONS
        self._actionsmod = None

    @property
    def actionsmod(self):
        if not self._actionsmod:
            astr = self.module.__name__ + '.actions'
            apath = astr.split('.')
            topmod = __import__(astr, globals(), locals(), [])
            mod = topmod
            for submod in apath[1:]:
                mod = getattr(mod, submod)
            self._actionsmod = mod
        return self._actionsmod


class ActionManager(object):
    """This class manages action modules and contained actions."""
    def __init__(self):
        self.actions = []
        self.modules = []

    def _actions_dict(self):
        adict = {}
        for action in self.actions:
            adict[action.name] = action.module
        return adict

    def add_actions_module(self, module, name=None, override=True):
        def _assign_module(action, override=False):
            if override or not action.module:
                action.module = name
            if action.steps:
                for step in action.steps:
                    _assign_module(step)

        def _override_action(action):
            for i, _action in enumerate(self.actions):
                if _action.name == action.name:
                    self.actions[i] = action
                    break

        if not name:
            _, _, name = module.__name__.rpartition('.')
        adict = self._actions_dict()
        if not hasattr(module, 'ACTIONS'):
            raise exception.InvalidActionModule(
                module=module.__name__, why="missing ACTIONS description")
        for action in module.ACTIONS:
            if action.name in adict:
                # existing action
                if override:
                    _assign_module(action, override=True)
                    log.debug("overriding %s (module: %s -> %s)" %
                              (action, adict[action.name], name))
                    _override_action(action)
                else:
                    raise exception.DuplicateAction(
                        action=action.name, mod1=adict[action.name], mod2=name)
            else:
                # new action
                _assign_module(action, override=True)
                self.actions.append(action)
        self.modules.append(ActionModule(module, name))

    def add_actions_modules(self, package, override=True):
        added = []
        for importer, modname, ispkg in pkgutil.iter_modules(package.__path__):
            modpath = '%s.%s' % (package.__name__, modname)
            if not ispkg:
                continue
            try:
                mod = importer.find_module(modname).load_module(modname)
            except ImportError:
                log.warn("Failed to import module: %s" % modpath)
                continue
            if not hasattr(mod, 'ACTIONS'):
                continue
            self.add_actions_module(mod, name=modname, override=override)
            added += modname
        return added

    def fill_aliases(self):
        def _fill_alias(dst, src):
            for attr in ['help',
                         'description',
                         'optional_args',
                         'required_args']:
                src_val = getattr(src, attr)
                if not getattr(dst, attr) and src_val:
                    setattr(dst, attr, src_val)

        for a in self.actions:
            if not a.alias:
                continue
            for b in self.actions:
                if a.alias == b.name:
                    _fill_alias(a, b)

    def action_str(self, action):
        return ".".join(map(lambda x: x.name, action))

    def action_serialize(self, action):
        return list(map(lambda x: x.name, action))

    def action_deserialize(self, action_sr):
        action = []
        steps = self.actions
        for action_name in action_sr:
            found = False
            for step in steps:
                if six.text_type(step.name) == action_name:
                    action.append(step)
                    steps = step.steps or []
                    found = True
                    break
            if not found:
                astr = self.action_str(action + [Action(action_name)])
                raise exception.InvalidAction(action=astr)
        return action

    def ensure_leaf_action(self, action, on_change_callback=None):
        step = action[-1]
        if step.steps:
            while step.steps:
                next_step = step.steps[0]
                action.append(next_step)
                step = next_step
            if on_change_callback:
                on_change_callback()
        return action

    def next_action(self, action):
        while action:
            last = action[-1]
            if len(action) <= 1:
                return None
            parent = action[-2]
            found = False
            for step in parent.steps:
                if step == last:
                    found = True
                elif found:
                    action = action[:-1] + [step]
                    return self.ensure_leaf_action(action)
            action = action[:-1]
        return None

    def _get_action_fun(self, action):
        fun = None
        for m in self.modules:
            if m.name != action.module:
                continue
            _fun = getattr(m.actionsmod, action.name, None)
            if _fun:
                fun = _fun
                break
        return fun

    def run_action(self, action, args=None):
        if not args:
            args = {}
        if action.continuable:
            log.info(log.term.bold("## %s" % action.name))
        for carg in action.const_args:
            args[carg] = action.const_args[carg]
        action_fun = action.action_fun
        if not action_fun:
            action_fun = self._get_action_fun(action)
            if not action_fun:
                raise exception.ActionFunctionNotAvailable(
                    action=action.name, module=action.module)
            action.action_fun = action_fun

        argspec = inspect_getargspec(action_fun)
        fun_args = []
        if argspec.defaults:
            n_defaults = len(argspec.defaults)
        else:
            n_defaults = 0
        n_required = len(argspec.args) - n_defaults
        for i, arg in enumerate(argspec.args):
            if arg in args:
                fun_args.append(args[arg])
            else:
                if i < n_required:
                    raise exception.RequiredActionArgumentNotAvailable(
                        action=action.name, arg=arg)
                else:
                    fun_args.append(argspec.defaults[i - n_required])
        return action_fun(*fun_args)

    def print_actions(self):
        for action in self.actions:
            print(str(action))



"""
Requirements related actions.
"""
from __future__ import print_function
import os
import re
import yaml

from rdopkg import exception
from rdopkg import guess
from rdopkg.actionmods import query as _query
from rdopkg.actionmods import reqs as _reqs
from rdopkg.utils import log
from rdopkg.utils.git import git


def reqdiff(version_tag_from, version_tag_to):
    fmt = "\n{t.bold}requirements.txt diff{t.normal} between " \
          "{t.bold}{old}{t.normal} and {t.bold}{new}{t.normal}:"
    log.info(fmt.format(t=log.term, old=version_tag_from, new=version_tag_to))
    rdiff = _reqs.reqdiff_from_refs(version_tag_from, version_tag_to)
    _reqs.print_reqdiff(*rdiff)


def reqcheck(version, spec=False):
    if version.upper() == 'XXX':
        if 'upstream' in git.remotes():
            current_branch = git.current_branch()
            branch = current_branch.replace('rpm-', '')
            if branch != 'master':
                branch = 'stable/{}'.format(branch)
            version = 'upstream/{}'.format(branch)
            check = _reqs.reqcheck_spec(ref=version)
        else:
            m = re.search(r'/([^/]+)_distro', os.getcwd())
            if not m:
                raise exception.CantGuess(what="requirements.txt location",
                                          why="failed to parse current path")
            path = '../%s/requirements.txt' % m.group(1)
            log.info("Delorean detected. Using %s" % path)
            check = _reqs.reqcheck_spec(reqs_txt=path)
    else:
        check = _reqs.reqcheck_spec(ref=version)
    format = None
    if spec:
        format = 'spec'
    _reqs.print_reqcheck(*check, format=format)


def reqquery(reqs_file=None, reqs_ref=None, spec=False, filter=None,
             dump=None, dump_file=None, load=None, load_file=None,
             verbose=False):
    if not (reqs_ref or reqs_file or spec or load or load_file):
        reqs_ref = guess.patches_base_ref()
    if not (bool(reqs_ref) ^ bool(reqs_file) ^ bool(spec)
            ^ bool(load) ^ bool(load_file)):
        raise exception.InvalidUsage(
            why="Only one requirements source (-r/-R/-s/-l/-L) can be "
                "selected.")
    if dump and dump_file:
        raise exception.InvalidUsage(
            why="Only one dump method (-d/-D) can be selected.")
    if dump:
        dump_file = 'requirements.yml'
    if load:
        load_file = 'requirements.yml'

    # get query results as requested
    if load_file:
        log.info("Loading query results from file: %s" % load_file)
        r = yaml.load(open(load_file))
    else:
        release, dist = None, None
        if not filter:
            try:
                release, dist = guess.osreleasedist()
                log.info('Autodetected filter: %s/%s'
                         % (release, dist))
            except exception.CantGuess as ex:
                raise exception.CantGuess(
                    msg='%s\n\nPlease select RELEASE(/DIST) filter to query.' %
                        str(ex))
        else:
            release, _, dist = filter.partition('/')
        module2pkg = True
        if reqs_file:
            log.info("Querying requirements file: %s" % reqs_file)
            reqs = _reqs.get_reqs_from_path(reqs_file)
        elif reqs_ref:
            log.info("Querying requirements file from git: "
                     "%s -- requirements.txt" % reqs_ref)
            reqs = _reqs.get_reqs_from_ref(reqs_ref)
        else:
            log.info("Querying .spec file")
            module2pkg = False
            reqs = _reqs.get_reqs_from_spec(as_objects=True)
        log.info('')
        r = _reqs.reqquery(reqs, release=release, dist=dist,
                           module2pkg=module2pkg, verbose=verbose)

    if dump_file:
        log.info("Saving query results to file: %s" % dump_file)
        yaml.dump(r, open(dump_file, 'w'))

    _reqs.print_reqquery(r)


def query(filter, package, verbose=False):
    r = _query.query_rdo(filter, package, verbose=verbose)
    if not r:
        log.warn('No distrepos information in rdoinfo for %s' % filter)
        return
    if verbose:
        print('')
    _query.pretty_print_query_results(r)



# -*- encoding: utf-8 -*-
from __future__ import print_function
import argparse

from rdopkg import exception
from rdopkg.utils import log


ARGCOMPLETE_AVAILABLE = False
try:
    import argcomplete
    ARGCOMPLETE_AVAILABLE = True
except ImportError:
    pass


def action2cmd(action):
    return action.replace('_', '-')


def optarg2cmdarg(arg):
    return "--%s" % action2cmd(arg)


def get_action_args(action, args):
    aargs = {}
    for rarg in action.required_args:
        aargs[rarg.name] = getattr(args, action2cmd(rarg.name))
    for oarg in action.optional_args:
        val = getattr(args, oarg.name)
        if val:
            aargs[oarg.name] = val
    return aargs


def get_parser(runner, prog='rdopkg', version=None):
    parser = argparse.ArgumentParser(
        prog=prog,
        formatter_class=argparse.RawTextHelpFormatter)
    subparsers = parser.add_subparsers(help='available actions')
    parser.add_argument('-c', '--continue', action='store_true',
                        help='continue running current action')
    if version:
        parser.add_argument('--version', action='version', version=version)
    for action in runner.action_manager.actions:
        cmd = action2cmd(action.name)
        action_parser = subparsers.add_parser(
            cmd,
            help=action.help,
            description=action.description or action.help,
            formatter_class=argparse.RawTextHelpFormatter)
        for oarg in action.optional_args:
            if oarg.positional:
                arg_names = [oarg.name]
            else:
                arg_names = oarg.cmdarg_names(optarg2cmdarg)
            action_parser.add_argument(*arg_names, **oarg.kwargs)
        for rarg in action.required_args:
            arg_names = rarg.cmdarg_names(action2cmd)
            action_parser.add_argument(*arg_names, **rarg.kwargs)
        action_parser.set_defaults(action=action)
    return parser


def run(action_runner, cargs, prog='rdopkg', version=None):
    parser = get_parser(action_runner, prog=prog, version=version)
    if ARGCOMPLETE_AVAILABLE:
        argcomplete.autocomplete(parser)

    code = 1
    try:
        if not cargs:
            parser.print_help()
            return
        # argparse still can't do this with subparsers (reported 2009)
        elif '--abort' in cargs:
            action_runner.clear_state(verbose=True)
            return 1
        elif '--continue' in cargs or cargs == ('-c',):
            action_runner.load_state_safe()
        else:
            args = parser.parse_args(cargs)
            action = args.action
            action_args = get_action_args(action, args)
            action_runner.new_action(action, action_args)

        action_runner.engage()
        code = 0
    except (
            exception.ActionInProgress,
            exception.BranchNotFound,
            exception.CantGuess,
            exception.ConfigError,
            exception.CommandNotFound,
            exception.DebugAbort,
            exception.FileNotFound,
            exception.IncompleteChangelog,
            exception.InvalidLintCheck,
            exception.InvalidPatchesBaseRef,
            exception.InvalidPackageFilter,
            exception.InvalidPackage,
            exception.InvalidReleaseBumpIndex,
            exception.InvalidGitRef,
            exception.InvalidQuery,
            exception.InvalidUsage,
            exception.LintProblemsFound,
            exception.ManualResolutionNeeded,
            exception.NoPatchesChanged,
            exception.NoDistgitChangesFound,
            exception.NotADirectory,
            exception.RequiredActionArgumentNotAvailable,
            exception.RequirementNotMet,
            exception.RpmModuleNotAvailable,
            exception.SpecFileNotFound,
            exception.ToolNotFound,
            exception.UserAbort,
            exception.UnverifiedPatch,
            KeyboardInterrupt,
    ) as ex:
        code = getattr(ex, 'exit_code', code)
        print("")
        print(log.term.important(str(ex) or type(ex).__name__))
    except exception.NoActionInProgress as ex:
        print("%s See `rdopkg -h` for help." % ex)
    except exception.CommandFailed as ex:
        # this was logged already
        pass
    except exception.InternalAction as ex:
        if ex.kwargs['action'] == 'status':
            action_runner.load_state_safe()
            action_runner.status()
            return 0
        elif ex.kwargs['action'] == 'actions':
            fails = action_runner.check_actions()
            return min(1, len(fails))
        raise
    return code



from __future__ import unicode_literals
from builtins import str

import json
import subprocess

from rdopkg import exception
from rdopkg.utils import log


class _CommandOutput(str):
    """
    Just a string subclass with attribute access.
    """
    @property
    def success(self):
        return self.return_code == 0


def log_cmd_fail(cmd, cout, fail_log_fun=log.warn, out_log_fun=log.info):
    fail_log_fun('{t.error}command failed: {t.normal}{t.cmd}{cmd}{t.normal}'
                 .format(t=log.term, cmd=cmd))
    nl = False
    if cout:
        out_log_fun(log.term.bold("stdout:"))
        out_log_fun(cout)
        nl = True
    if cout.stderr:
        out_log_fun(log.term.bold("stderr:"))
        out_log_fun(cout.stderr)
        nl = True
    if nl:
        out_log_fun('')


def run(cmd, *params, **kwargs):
    fatal = kwargs.get('fatal', True)
    direct = kwargs.get('direct', False)
    log_cmd = kwargs.get('log_cmd', True)
    log_fail = kwargs.get('log_fail', True)
    input = kwargs.get('input')
    print_stdout = kwargs.get('print_stdout', False)
    print_stderr = kwargs.get('print_stderr', False)
    print_output = kwargs.get('print_output', False)
    env = kwargs.get('env', None)

    cmd = [cmd] + list(params)
    cmd_str = ' '.join(cmd)

    if log_cmd:
        log.command(log.term.cmd(cmd_str))

    if print_output:
        print_stdout = True
        print_stderr = True

    if input:
        stdin = subprocess.PIPE
        input = input.encode()
    else:
        stdin = None

    if direct:
        stdout = None
        stderr = None
    else:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE

    try:
        prc = subprocess.Popen(cmd, stdin=stdin, stdout=stdout,
                               stderr=stderr, env=env)
    except OSError:
        raise exception.CommandNotFound(cmd=cmd[0])
    out, err = prc.communicate(input=input)

    if out:
        out = out.rstrip()
        if print_stdout:
            log.info(out)
    else:
        out = b''

    if err:
        err = err.rstrip()
        if print_stderr:
            log.info(err)
    else:
        err = b''

    cout = _CommandOutput(out.decode('utf-8'))
    cout.stderr = err
    cout.return_code = prc.returncode
    cout.cmd = cmd_str
    if prc.returncode != 0:
        if log_fail:
            log_cmd_fail(cmd_str, cout)
        if fatal:
            raise exception.CommandFailed(cmd=cmd, out=cout)
    return cout


class ShellCommand(object):
    command = None

    def __init__(self):
        if self.command is None:
            self.command = self.__class__.__name__.lower()

    def __call__(self, *params, **kwargs):
        return run(self.command, *params, **kwargs)


class GerritQuery(ShellCommand):
    def __init__(self, host, port, log_cmd=True):
        self.host = host
        self.port = port
        self.log_cmd = log_cmd

    def __call__(self, *params, **kwargs):
        if 'log_cmd' not in kwargs:
            kwargs['log_cmd'] = self.log_cmd
        results = run('ssh', '-p', self.port, self.host,
                      'gerrit', 'query', '--format=JSON',
                      *params, **kwargs)
        # gerrit sends stats that we need to discard
        result = results.split('}\n{')[0]
        try:
            j = json.loads(result + '}')
        except ValueError:
            j = json.loads(result)
        if j.get('type') in ['error', 'stats']:
            # we have no results, just the query stats
            return None
        else:
            return j



import contextlib
import os
import re

from rdopkg import exception
from rdopkg.utils.cmd import run
from rdopkg.utils.cmd import ShellCommand
from rdopkg.utils.issues import search_bug_references


@contextlib.contextmanager
def git_branch(branch):
    if branch:
        old_branch = git.current_branch()
        git('checkout', branch)
        try:
            yield
        finally:
            git('checkout', old_branch)
    else:
        yield


@contextlib.contextmanager
def git_revision(gitrev):
    if gitrev:
        old_gitrev = git.current_commit()
        git.checkout(gitrev)
        try:
            yield
        finally:
            git.checkout(old_gitrev)
    else:
        yield


class Git(ShellCommand):
    command = "git"

    def __call__(self, *params, **kwargs):
        # allows us to run git in isolated mode, avoiding interaction with user
        # specific config, like ~/.git-templates/hooks (used for testing)
        if kwargs.get('isolated', False):
            env = kwargs.get('env', os.environ.copy())
            env['GIT_CONFIG_NOSYSTEM'] = '1'
            env['GIT_CONFIG_NOGLOBAL'] = '1'
            kwargs['env'] = env
        return run(self.command, *params, **kwargs)

    def create_branch_from_remote(self, branch, remote_branch=None):
        lbr = self.local_branches()
        if branch in lbr or remote_branch in lbr:
            return
        if remote_branch is None:
            for rbr in self.remote_branches():
                br = rbr[rbr.find('/') + 1:]
                if rbr == branch or br == branch:
                    remote_branch = rbr
        elif remote_branch not in self.remote_branches():
            raise Exception("Branch %s doesn't exist in remotes" %
                            remote_branch)
        self.create_branch(branch, remote_branch)

    def _parse_output(self, out):
        output = [l.strip() for l in out.split("\n") if l]
        return output

    def _parse_branch_output(self, out):
        output = [l for l in self._parse_output(out) if l.find('HEAD') < 0]
        return output

    def remote_branches(self, remote=""):
        res = self("branch", "-r", "--no-color")
        branches = self._parse_branch_output(res)
        branches = [b.replace("remotes/", "")
                    for b in branches if b.startswith(remote)]
        return branches

    def local_branches(self):
        res = self("branch")
        res = self._parse_branch_output(res)
        res = [b.replace("* ", "") for b in res]
        return res

    def current_branch(self):
        branch = self('rev-parse', '--abbrev-ref', 'HEAD', log_cmd=False)
        return branch

    def current_commit(self):
        commit = self('rev-parse', 'HEAD', log_cmd=False)
        return commit

    def current_commit_message(self):
        commit_msg = self('log', '-n1', '--pretty=%B', log_cmd=False)
        return commit_msg

    def ref_exists(self, ref):
        o = self('show-ref', '--verify', '--quiet', ref,
                 fatal=False, log_cmd=False, log_fail=False)
        return o.success

    def branch_exists(self, branch):
        return self.ref_exists('refs/heads/%s' % branch)

    def object_type(self, ref):
        o = self('cat-file', '-t', ref,
                 fatal=False, log_cmd=False, log_fail=False)
        if not o:
            return None
        return o

    def is_clean(self):
        o = self('status', '-uno', '--porcelain', log_cmd=False)
        if o:
            return False
        return True

    def remotes(self):
        res = self("remote", "show", log_cmd=False)
        return self._parse_branch_output(res)

    def remote_branch_split(self, branch, fatal=True):
        parts = branch.split('/')
        n_parts = len(parts)
        if n_parts >= 2:
            remotes = self.remotes()
            for n in range(1, n_parts):
                remote = '/'.join(parts[:n])
                if remote in remotes:
                    _branch = '/'.join(parts[n:])
                    return remote, _branch
        if fatal:
            raise exception.InvalidRemoteBranch(branch=branch)
        return None, branch

    def remote_of_local_branch(self, branch):
        try:
            o = self('for-each-ref', '--format=%(upstream:short)',
                     'refs/heads/%s' % branch, log_cmd=False, log_fail=False)
        except exception.CommandFailed:
            return None
        if not o:
            return None
        return o

    def delete_branch(self, branch):
        if self.branch_exists(branch):
            self('branch', '-D', branch)

    def create_branch(self, new_branch, branch):
        try:
            self('branch', '-f', new_branch, branch)
        except exception.CommandFailed:
            # this could only fail if we're on the branch
            self('reset', '--hard', branch)

    def squash_last(self, branch=None):
        if branch is not None:
            self('checkout', branch)
        self("reset", "--soft", "HEAD~")
        # on git < 1.7.9 --no-edit can be done by:
        # git commit -C HEAD --amend
        self("commit", "--amend", "--no-edit")

    def branch_needs_push(self, branch=None):
        if not branch:
            branch = self.current_branch()
        remote_branch = self.remote_of_local_branch(branch)
        h1 = self.get_latest_commit_hash(branch)
        h2 = self.get_latest_commit_hash(remote_branch)
        return h1 != h2

    def linearize(self, starting_point, branch=None):
        if branch is not None and self.current_branch() != branch:
            self('checkout', branch)
        self('rebase', starting_point)

    def rev_range(self, from_revision, to_revision=None):
        rng = from_revision + '..'
        if to_revision:
            rng += to_revision
        return rng

    def get_commits(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%h %s', rng, log_cmd=False)
        commits = self._parse_output(log_out)
        return map(lambda x: tuple(x.split(' ', 1)), commits)

    def get_commit_subjects(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%s', rng, log_cmd=False)
        return self._parse_output(log_out)

    def get_commit_hashes(self, from_revision, to_revision=None):
        rng = self.rev_range(from_revision, to_revision)
        log_out = self('log', '--format=%h', rng, log_cmd=False)
        return self._parse_output(log_out)

    def get_commit_bzs(self, from_revision, to_revision=None):
        """
        Return a list of tuples, one per commit. Each tuple is (sha1, subject,
        bz_list). bz_list is a (possibly zero-length) list of numbers.
        """
        rng = self.rev_range(from_revision, to_revision)
        GIT_COMMIT_FIELDS = ['id', 'subject', 'body']
        GIT_LOG_FORMAT = ['%h', '%s', '%b']
        GIT_LOG_FORMAT = '%x1f'.join(GIT_LOG_FORMAT) + '%x1e'
        log_out = self('log', '--format=%s' % GIT_LOG_FORMAT, rng,
                       log_cmd=False, fatal=False)
        if not log_out:
            return []

        log = log_out.strip('\n\x1e').split("\x1e")
        log = [row.strip('\n\t ').split("\x1f") for row in log]
        log = [dict(zip(GIT_COMMIT_FIELDS, row)) for row in log]
        result = []
        for commit in log:
            bzs = search_bug_references(commit['subject'])
            bzs.extend(search_bug_references(commit['body']))
            result.append((commit['id'], commit['subject'], bzs))
        return result

    def get_latest_commit_hash(self, ref=None):
        cmd = ['log', '-n', '1', '--format=%H']
        if ref:
            cmd.append(ref)
        out = self(*cmd, log_cmd=False, log_fail=False, fatal=False)
        return out

    def get_latest_tag(self, branch=None):
        cmd = ['describe', '--abbrev=0', '--tags']
        if branch:
            cmd.append(branch)
        out = self(*cmd, log_cmd=False)
        return out

    def get_file_authors(self, path, with_email=True):
        if with_email:
            pf = '%an <%ae>'
        else:
            pf = '%an'
        authors = self('log', '--oneline', "--pretty=%s" % pf, path,
                       log_cmd=False)
        authors = reversed(authors.split('\n'))
        return authors

    def get_file_content(self, rev, path):
        obj = '%s:%s' % (rev, path)
        return self('show', obj, log_cmd=False)

    def config_get(self, param, default=None):
        '''Return the value of a git configuration option.  This will
        return the value of the default parameter (which defaults to
        None) if the given option does not exist.'''

        try:
            return self("config", "--get", param,
                        log_fail=False, log_cmd=False)
        except exception.CommandFailed:
            return default

    def config_set(self, param, value, is_global=False):
        params = [param, value]
        if is_global:
            params.insert(0, '--global')
        return self("config", *params)

    def checkout(self, branch):
        self("checkout", branch, log_cmd=False)

    def remove(self, hash):
        self('rebase', '--onto', hash + '^', hash, '--preserve-merges',
             '--committer-date-is-author-date')


git = Git()



"""
Check for common .spec problems
"""
from rdopkg.utils import specfile
from rdopkg.utils import lint as _lint


def lint(spec_fn=None, lint_checks=None, error_level='E'):
    if not spec_fn:
        spec_fn = specfile.spec_fn()
    if lint_checks:
        checks = lint_checks.split(',')
    else:
        checks = None
    hints = _lint.lint(spec_fn, checks=checks)
    _lint.lint_report(hints, error_level=error_level)



"""
review workflow actions (rpmfactory, review.rdoproject.org)
"""

from rdopkg import guess
from rdopkg.actionmods import rpmfactory
from rdopkg.utils.git import git


def review_patch(local_patches_branch=None):
    if not local_patches_branch:
        local_patches_branch = git.current_branch()
        if not local_patches_branch.endswith('-patches'):
            br = guess.patches_branch(local_patches_branch)
            if br:
                local_patches_branch = br.partition('/')[2]
    rpmfactory.review_patch(local_patches_branch)


def review_spec(branch):
    rpmfactory.review_spec(branch)



from __future__ import print_function
from past.builtins import basestring
import distroinfo
import distroinfo.query
from distroinfo.info import DistroInfo

from rdopkg import exception, helpers
from rdopkg.utils import log
from rdopkg.conf import cfg


def print_releases(info):
    print("{t.bold}releases & repos:{t.normal}".format(t=log.term))
    for rls in info['releases']:
        s = "  {t.bold}{rls}{t.normal}".format(t=log.term, rls=rls['name'])
        if 'fedora' in rls:
            s += '   (Fedora %s)' % rls['fedora']
        print(s)
        for repo in rls['repos']:
            if 'special' in repo:
                print("    {t.bold}{name}{t.normal}: "
                      "{t.yellow}{special}{t.normal}".format(
                          t=log.term,
                          name=repo['name'],
                          special=repo['special']))
            else:
                print("    {t.bold}{name}{t.normal} built in"
                      " {t.bold}{bs}{t.normal} from"
                      " {t.bold}{branch}{t.normal} branch".format(
                          t=log.term,
                          name=repo['name'],
                          bs=repo.get('buildsys', '??'),
                          branch=repo['branch']))


def print_pkg_summary(info):
    pkgs = info['packages']
    n = len(pkgs)
    print("{t.bold}{n} packages defined:{t.normal}".format(
        t=log.term, n=n))
    confs = {}
    for pkg in pkgs:
        conf = pkg.get('conf')
        confs[conf] = confs.get(conf, 0) + 1

    for conf, n in confs.items():
        if not conf:
            continue
        print("  {t.bold}{n}{t.normal} using {t.bold}{conf}{t.normal} "
              "config".format(t=log.term, conf=conf, n=n))

    n = confs.get(None)
    if n:
        print("  {t.bold}{n}{t.normal} configured ad-hoc".format(
              t=log.term, n=n))


def print_summary(info):
    print_releases(info)
    print('')
    print_pkg_summary(info)


def print_pkgs(info, filters=None):
    pkgs = info['packages']
    if filters:
        pkgs = distroinfo.query.filter_pkgs(pkgs, rexen=filters)
    if not pkgs:
        print("No packages match your filter.")
        return
    print("{t.bold}{n} packages found:{t.normal}".format(t=log.term,
                                                         n=len(pkgs)))
    for pkg in pkgs:
        print("")
        print_pkg(pkg)


def print_pkg(pkg):
    dp = helpers.DictPrinter(
        header='name',
        first=['project', 'conf', 'upstream', 'patches', 'distgit', 'master',
               'distgit'],
        last=['maintainers'])
    dp(pkg)


def info_file(distro=None):
    """Return default distroinfo info file"""
    if not distro:
        distro = cfg['DISTRO']
    info_file_conf = distro.upper() + 'INFO_FILE'
    try:
        return cfg[info_file_conf]
    except KeyError:
        raise exception.InvalidUsage(
            why="Couldn't find config option %s for distro: %s"
                % (info_file_conf, distro))


def get_distroinfo(distro=None):
    """Get DistroInfo initialized from configuration"""
    if not distro:
        distro = cfg['DISTRO']
    _info_file = info_file(distro)
    # prefer git fetcher if available
    git_info_url_conf = distro.upper() + 'INFO_REPO'
    try:
        remote_git_info = cfg[git_info_url_conf]
        return DistroInfo(_info_file, remote_git_info=remote_git_info)
    except KeyError:
        pass
    # try raw remote fetcher
    remote_info_url_conf = distro.upper() + 'INFO_RAW_URL'
    try:
        remote_info = cfg[remote_info_url_conf]
        return DistroInfo(_info_file, remote_info=remote_info)
    except KeyError:
        raise exception.InvalidUsage(
            why="Couldn't find config option %s or %s for distro: %s"
                % (git_info_url_conf, remote_info_url_conf, distro))


def get_rdoinfo():
    """Compat function
    """
    return get_distroinfo(distro='rdo')


def get_default_inforepo(apply_tag=None, include_fns=None):
    raise DeprecationWarning("rdopkg >= 0.47.0 uses distroinfo, please use "
                             "`get_distroinfo` instead.")



import pytest

from rdopkg.cli import rdopkg, rdopkg_runner
from rdopkg import exception

import actions


def test_actions_availability():
    r = rdopkg('actions')
    assert r == 0, "Some action functions are NOT AVAILABLE"


def test_actions_continue_short():
    r = rdopkg('-c')
    assert r == 1, "-c succeeded with no state"


def test_actions_continue_long():
    r = rdopkg('--continue')
    assert r == 1, "--continue succeeded with no state"


def test_duplicate_action():
    runner = rdopkg_runner()
    aman = runner.action_manager
    with pytest.raises(exception.DuplicateAction):
        aman.add_actions_modules(actions, override=False)


def test_action_override():
    runner = rdopkg_runner()
    aman = runner.action_manager
    aman.add_actions_modules(actions, override=True)
    found = False
    for action in aman.actions:
        if action.name == 'clone':
            # make sure clone action is overridden with the custom one
            assert action.module == 'foomod'
            found = True
            break
    assert found, "clone action not found ?!"



# coding=utf-8
from rdopkg import exception
from rdopkg.utils import specfile
import pytest

import test_common as common


@pytest.mark.parametrize('version,numeric,rest', [
    ('', '', ''),
    ('1', '1', ''),
    ('1.2.3', '1.2.3', ''),
    ('1.2.3.0b1', '1.2.3', '.0b1'),
    ('10.10.10.0rc2', '10.10.10', '.0rc2'),
    ('a', 'a', ''),
    ('a.b', 'a.b', ''),
    ('1.b42', '1', '.b42'),
    ('1.0.b4.2', '1.0', '.b4.2'),
    ('1.2.3.a.b-c.d', '1.2.3', '.a.b-c.d'),
])
def test_version_parts(version, numeric, rest):
    parts = specfile.version_parts(version)
    assert parts == (numeric, rest)


@pytest.mark.parametrize('release,nums,milestone,macros', [
    ('1', '1', '', ''),
    ('0.1.2', '0.1.2', '', ''),
    ('1.b1', '1', '.b1', ''),
    ('0.1.rc2', '0.1', '.rc2', ''),
    ('0.1%{?dist}', '0.1', '', '%{?dist}'),
    ('0.1.b1%{?dist}', '0.1', '.b1', '%{?dist}'),
    ('0.1.b1.%{?dist}', '0.1', '.b1', '.%{?dist}'),
    ('0.1%{m1}%{m2}', '0.1', '', '%{m1}%{m2}'),
    ('0.1.%m1%m2', '0.1', '', '.%m1%m2'),
    ('0.1.0rc1%{m}', '0.1', '.0rc1', '%{m}'),
    ('0.1.0.0b2%m', '0.1.0', '.0b2', '%m'),
    ('0.1.0.000a000', '0.1.0', '.000a000', ''),
    ('10.10.10.%{?milestone}', '10.10.10', '.%{?milestone}', ''),
    ('10.10.10.%{?milestone}%{foo}',
     '10.10.10', '.%{?milestone}', '%{foo}'),
    ('0.0.0%{?milestone}.%bar',
     '0.0.0', '%{?milestone}', '.%bar'),
    ('%{ver}', '', '', '%{ver}'),
])
def test_release_parts(release, nums, milestone, macros):
    parts = specfile.release_parts(release)
    assert parts == (nums, milestone, macros)


@pytest.mark.parametrize('vr,epoch_arg,result', [
    ((None, '1.2.3', '0.1'), None, '1.2.3-0.1'),
    ((None, '1.2.3', '666%{?dist}'), False, '1.2.3-666'),
    ((None, '1.2.3', ''), True, '0:1.2.3'),
    ((23, '1.2.3', ''), None, '23:1.2.3'),
    ((23, '1.2.3', '0.1'), False, '1.2.3-0.1'),
    ((23, '1.2.3', '666%{?dist}'), True, '23:1.2.3-666'),
])
def test_get_vr(vr, epoch_arg, result):
    epoch, version, release = vr

    def _get_tag_mock(tag, default=None, expand_macros=False):
        if tag == 'Version':
            return version
        elif tag == 'Release':
            return release
        elif tag == 'Epoch':
            if epoch:
                return epoch
            raise exception.SpecFileParseError(
                spec_fn='TESTING', error='Pretending Epoch tag not found')
        return "MOCKED-OUT-VR"

    def _expand_macro(macro):
        return macro

    spec = specfile.Spec()
    spec.get_tag = _get_tag_mock
    spec.expand_macro = _expand_macro
    vr = spec.get_vr(epoch=epoch_arg)
    assert vr == result


def test_patches_base_add_patched(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        ver, np = spec.get_patches_base()
        assert ver is None
        assert np == 0
        spec.set_patches_base('+2')
        spec.save()
    common.assert_distgit(dist_path, 'patched-ex')


def test_patches_base_add_empty(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        ver, np = spec.get_patches_base()
        assert ver is None
        assert np == 0
        spec.set_patches_base_version('+2')
        spec.save()
    common.assert_distgit(dist_path, 'empty-ex')


def test_patches_base_noop_weird(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-weird')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        ver, np = spec.get_patches_base()
        assert ver == 'banana'
        assert np == 2
        spec.set_patches_base_version('banana')
        spec.save()
    common.assert_distgit(dist_path, 'empty-weird')


def test_set_commit_ref_macro(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'commit')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        spec.set_commit_ref_macro('86a713c35718520cb3b681260182a8388ac809f3')
        spec.save()
    common.assert_distgit(dist_path, 'commit-patched')


def test_get_patches_ignore_regex(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-ex-filter')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        regex = spec.get_patches_ignore_regex()
        assert regex.pattern == 'DROP-IN-RPM'


def test_get_patches_ignore_regex_fail(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-filter-bogus')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        regex = spec.get_patches_ignore_regex()
        assert regex is None


def test_new_version_patches_base_ignore_load(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'magic-comments')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        assert spec.get_patches_ignore_regex().pattern == 'DROP-IN-RPM'
        assert spec.get_patches_base() == ('1.2.3', 0)


def test_new_version_patches_base_ignore_new_version(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'magic-comments')
    with dist_path.as_cwd():
        spec2 = specfile.Spec()
        spec2.set_tag('Version', '1.2.4')
        spec2.set_release('1')
        spec2.set_patches_base_version(None)
        spec2.save()
        spec2 = None
        spec = specfile.Spec()
        assert spec.get_patches_ignore_regex().pattern == 'DROP-IN-RPM'


def test_new_version_patches_base_ignore_mangling_minimal(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'magic-comments')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        spec.set_patches_base_version(None)
        assert spec.get_patches_ignore_regex().pattern == 'DROP-IN-RPM'


def test_new_version_patches_base_ignore_mangling_minimal_2(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched-filter')
    with dist_path.as_cwd():
        spec = specfile.Spec()
        spec.set_patches_base_version(None)
        assert spec.get_patches_ignore_regex().pattern == 'DROP-IN-RPM'


def test_set_patches_base_case_1_minimal():
    spec = specfile.Spec(txt='foo\n# patches_base=1.2.3\nbaz\n')
    spec.set_patches_base_version(None)
    assert spec.txt == 'foo\nbaz\n'


def test_set_patches_base_case_1_a_minimal():
    spec = specfile.Spec(txt='foo\n#patches_base=1.2.3\nbaz\n')
    spec.set_patches_base_version(None)
    assert spec.txt == 'foo\nbaz\n'


# make sure dropping patches_base with following # is clean
def test_set_patches_base_case_2_minimal():
    spec = specfile.Spec(txt='foo\n# patches_base=1.2.3\n#\nbaz\n')
    spec.set_patches_base_version(None)
    assert spec.txt == 'foo\nbaz\n'


# found bug where following comment mangles prior line when clearing
# patches_base
def test_set_patches_base_case_3_minimal():
    spec = specfile.Spec(txt='foo\n# patches_base=1.2.3\n# bar\nbaz\n')
    spec.set_patches_base_version(None)
    assert 'foo\n' in spec.txt
    assert 'patches_base' not in spec.txt
    assert 'baz\n' in spec.txt


# make sure multiple whitespace leaves Version line intact (workaround)
def test_set_patches_base_case_4_minimal():
    spec = specfile.Spec(txt='foo\n\n# patches_base=1.2.3\n# bar\nbaz\n')
    spec.set_patches_base_version(None)
    assert 'foo\n' in spec.txt
    assert 'patches_base' not in spec.txt
    assert 'baz\n' in spec.txt


# Eat Trailing #\n comments
def test_set_patches_base_case_5_minimal():
    spec = specfile.Spec(txt='foo\n\n# patches_base=1.2.3\n#\n#\nbaz\n')
    spec.set_patches_base_version(None)
    assert 'foo\n' in spec.txt
    assert '^#\n' not in spec.txt
    assert 'baz\n' in spec.txt


# patches_base/patches_ignore
def test_set_patches_base_case_6_minimal():
    spec = specfile.Spec(txt='Version: 1.2.3\n\n# patches_base=1.2.3\n# patches_ignore=DROP-IN-RPM\n#\nPatch0=foo.patch\n')  # noqa
    spec.set_patches_base_version(None)
    assert 'Version: 1.2.3\n' in spec.txt
    assert '# patches_ignore=DROP-IN-RPM\n' in spec.txt
    assert 'Patch0=foo.patch\n' in spec.txt


# patches_ignore/patches_base
def test_set_patches_base_case_7_minimal():
    spec = specfile.Spec(txt='Version: 1.2.3\n\n# patches_ignore=DROP-IN-RPM\n# patches_base=1.2.3\n#\nPatch0=foo.patch\n')  # noqa
    spec.set_patches_base_version(None)
    assert 'Version: 1.2.3\n' in spec.txt
    assert '# patches_ignore=DROP-IN-RPM\n' in spec.txt
    assert 'Patch0=foo.patch\n' in spec.txt


# patches_ignore/patches_base keep patches_base
def test_set_patches_base_case_8_minimal():
    spec = specfile.Spec(txt='Version: 1.2.3\n\n# patches_ignore=DROP-IN-RPM\n# patches_base=\n#\nPatch0=foo.patch\n')  # noqa
    spec.set_patches_base_version(None)
    assert '# patches_ignore=DROP-IN-RPM\n' in spec.txt
    assert '# patches_base=1.2.3\n' in spec.txt
    assert '# patches_ignore=DROP-IN-RPM\n' in spec.txt


def test_get_magic_comment(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched-filter')
    with dist_path.as_cwd():
        spec = specfile.Spec()

        def _assert_mc(name, exp_val):
            spec_val = spec.get_magic_comment(name)
            assert spec_val == exp_val

        _assert_mc('patches_base', '+2')
        _assert_mc('patches_ignore', 'DROP-IN-RPM')
        _assert_mc('not_really_there', None)
        _assert_mc('¯\_(ツ)_/¯', None)


def test_get_magic_comment_minimal_1():
    spec = specfile.Spec(txt='\n')
    assert spec.get_magic_comment('foo') is None


def test_get_magic_comment_minimal_2():
    spec = specfile.Spec(txt='# foo=1\n')
    assert '1' == spec.get_magic_comment('foo')


def test_get_magic_comment_minimal_2a():
    spec = specfile.Spec(txt='#foo=1 \n')
    assert '1' == spec.get_magic_comment('foo')


def test_get_magic_comment_minimal_2b():
    spec = specfile.Spec(txt='#     foo = 1\n')
    assert '1' == spec.get_magic_comment('foo')


def test_get_magic_comment_minimal_3():
    spec = specfile.Spec(txt='# foo\n')
    assert spec.get_magic_comment('foo') is None


def test_get_magic_comment_minimal_4():
    spec = specfile.Spec(txt='# foo=1\n# bar=baz\n')
    assert 'baz' == spec.get_magic_comment('bar')


def test_get_magic_comment_minimal_5():
    spec = specfile.Spec(txt='#\n#\n# foo=1\n#\n# bar=baz\n#\n\n')
    assert 'baz' == spec.get_magic_comment('bar')


def test_get_magic_comment_minimal_6():
    spec = specfile.Spec(txt='  # foo=1\n')
    assert spec.get_magic_comment('foo') is None


@pytest.mark.skip(
    'Ignoring eged case until we get more details for expected behavior')
def test_get_magic_comment_minimal_7():
    spec = specfile.Spec(txt='# foo=\n')
    assert '' == spec.get_magic_comment('foo')


def test_get_last_changelog_entry_0():
    txt = 'Version: 1.2.3\n\n# patches_ignore=DROP-IN-RPM\n# patches_base=1.2.3\n#\nPatch0=foo.patch\n'  # noqa
    spec = specfile.Spec(txt=txt)
    r = spec.get_last_changelog_entry()
    assert ('', []) == r


def test_get_last_changelog_entry_1():
    txt = 'Version: 1.2.3\n\n# patches_ignore=DROP-IN-RPM\n# patches_base=1.2.3\n#\nPatch0=foo.patch\n'  # noqa
    spec = specfile.Spec(txt=txt + '%changelog\nfoo')
    r = spec.get_last_changelog_entry()
    assert ('foo', []) == r


def test_get_last_changelog_entry_1_case_insensitive():
    txt = 'Version: 1.2.3\n\n# patches_ignore=DROP-IN-RPM\n# patches_base=1.2.3\n#\nPatch0=foo.patch\n'  # noqa
    spec = specfile.Spec(txt=txt + '%ChangeLog\nfoo')
    r = spec.get_last_changelog_entry()
    assert ('foo', []) == r


def test_get_last_changelog_entry_multiple_sections():
    with pytest.raises(exception.MultipleChangelog):
        txt = 'Version: 1.2.3\n\n'
        spec = specfile.Spec(txt=txt + '%changelog\n* 2017-01-01\n- foo1\n\n%changelog\nbar\n')  # noqa
        r = spec.get_last_changelog_entry()
        assert False, r



import re


def search_bug_references(text_blob):
    """Return list of bug refs found for provided txt"""

    if text_blob is None:
        return []

    return re.findall(r'rhbz#(\d+)', text_blob)



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('review_patch',
           help="send patch(es) for review",
           optional_args=[
               Arg('local_patches_branch', metavar='PATCHES_BRANCH',
                   positional=True, nargs='?',
                   help="local patches branch with changes to review"),
           ]),
    Action('review_spec',
           help="send distgit (.spec file) change for review",
           optional_args=[
               Arg('branch', metavar='DISTGIT_BRANCH',
                   positional=True, nargs='?',
                   help="local distgit branch with changes to review"),
           ],
           steps=[
               Action('get_package_env', module='distgit'),
               Action('review_spec'),
           ]),
]



"""
Check for common .spec problems
"""

from collections import defaultdict
import re

from rdopkg import exception
from rdopkg.utils import log
from rdopkg.utils.cmd import run
from rdopkg import helpers


class LintHint(object):
    def __init__(self, location, level, msg):
        self.location = location
        self.level = level
        self.msg = msg

    def __repr__(self):
        return ('%s: %s: %s' %
                (self.location, self.level, self.msg))

    def pretty_str(self):
        fmt = "{loc}: {t.%s}{lvl}{t.normal}: {t.bold}{msg1}{t.normal} {msg2}"
        if self.level == 'W':
            color = 'yellow'
        else:
            color = 'red'
        fmt = fmt % color
        msg1, _, msg2 = self.msg.partition(' ')
        return fmt.format(
            loc=self.location,
            lvl=self.level,
            msg1=msg1,
            msg2=msg2,
            t=log.term
        )


RE_RPMLINT_HINT = r'(.*):\s+([EW]):\s+(.+)$'
RE_RPMLINT_SUMMARY = (r'\d+ packages and \d+ specfiles checked; '
                      r'\d+ errors?, \d+ warnings?.')


def rpmlint_check(*args):
    # run rpmlint and return linting hints it found
    hints = []
    cmd = ['rpmlint']
    if args:
        cmd += args
    try:
        out = run(*cmd, fatal=False, log_fail=False)
    except exception.CommandNotFound:
        raise exception.CommandNotFound(
            msg="Unable to run rpmlint checks because rpmlint is missing.")
    for line in out.splitlines():
        m = re.match(RE_RPMLINT_HINT, line)
        if m:
            hints.append(LintHint(location=m.group(1),
                                  level=m.group(2),
                                  msg=m.group(3)))
            continue
        m = re.match(RE_RPMLINT_SUMMARY, line)
        if m:
            # ignore final rpmlint summary
            continue
        hints.append(LintHint(location='rpmlint', level='W', msg=(
            'Failed to parse rpmlint output: %s' % line)))
    return hints


def sanity_check_buildarch(txt):
    # make sure BuildArch is AFTER SourceX and PatchX lines,
    # otherwise %{patches} macro is empty which causes trouble
    bm = re.search(r'^BuildArch:', txt, flags=re.M)
    if not bm:
        return True
    bi = bm.start()
    sm = re.search(r'^Source\d+:', txt, flags=re.M)
    if sm:
        si = sm.start()
        if bi < si:
            return False
    pm = re.search(r'^Patch\d+:', txt, flags=re.M)
    if pm:
        pi = pm.start()
        if bi < pi:
            return False
    return True


def sanity_check_patches_base(txt):
    # duplicate patches_base might lead to unexpected behavior
    bases = re.findall(r'^#\s*patches_base', txt, flags=re.M)
    if len(bases) > 1:
        return False
    return True


def sanity_check(spec_fn):
    # perform rdopkg sanity checks for common problems
    hints = []
    try:
        txt = open(spec_fn, 'r').read()
    except Exception as e:
        hints.append(LintHint(
            spec_fn, 'E', str(e)))
        return hints
    if not sanity_check_buildarch(txt):
        hints.append(LintHint(spec_fn, 'E', (
            "buildarch-before-sources: Due to mysterious"
            "ways of rpm, BuildArch needs to be placed "
            "AFTER SourceX and PatchX lines in .spec file, "
            "otherwise %{patches} macro will be empty "
            "and both %autosetup and `git am %{patches}` will fail. "
            "Please move BuildArch AFTER SourceX and PatchX lines.")))
    if not sanity_check_patches_base(txt):
        hints.append(LintHint(spec_fn, 'E', (
            "duplicate-patches-base: Please make sure to only have one "
            "# patches_base= entry in .spec file to avoid problems.")))
    return hints


LINT_CHECKS = {
    'sanity': sanity_check,
    'rpmlint': rpmlint_check,
}


def lint(spec_fn, checks=None):
    if not checks or checks == 'all':
        # use all checks by default
        checks = LINT_CHECKS.keys()
    hints = []
    for ch in checks:
        try:
            check = LINT_CHECKS[ch]
        except KeyError:
            raise exception.InvalidLintCheck(check=ch)
        hints += check(spec_fn)
    return hints


def lint_report(hints, error_level=None):
    # print a linting report based on passed hints
    # optionally raise error depending on error_level:
    #   * 'E': raise when errors found
    #   * 'W': raise when errors or warnings are found
    #   * other: don't raise error, only print report
    hint_dict = defaultdict(list)
    for hint in hints:
        hint_dict[hint.level].append(hint)

    errors = hint_dict.get('E', [])
    n_errors = len(errors)
    if errors:
        print("\n{t.red}{n} ERRORS found:{t.normal}".format(
            n=n_errors, t=log.term))
        for e in errors:
            print(e.pretty_str())

    warns = hint_dict.get('W', [])
    n_warns = len(warns)
    if warns:
        print("\n{t.yellow}{n} WARNINGS found:{t.normal}".format(
            n=n_warns, t=log.term))
        for e in warns:
            print(e.pretty_str())

    if n_errors == 0 and n_warns == 0:
        print("\n{t.green}no linting errors found \\o/{t.normal}".format(
            t=log.term))
        return

    critical_error = False
    if error_level == 'E' and n_errors > 0:
        critical_error = True
    if error_level == 'W' and (n_errors > 0 or n_warns > 0):
        critical_error = True
    if critical_error:
        raise exception.LintProblemsFound(
            "Linting problems detected: %d errors, %d warnings" % (
                n_errors, n_warns))



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('lint', help="check for common .spec problems",
           optional_args=[
               Arg('spec_fn', positional=True, nargs='?',
                   help="a .spec file to check"),
               Arg('lint_checks', metavar='CHECKS',
                   help="comma-separated lists of checks to perform "
                        "(checks: sanity, rpmlint, all)"),
               Arg('error_level', metavar='E|W|-',
                   help="'E' to halt on errors (default), "
                        "'W' to halt on errors and warnings, "
                        "'-' to ignore errors (affects return value)")
           ])
]



import pytest
import subprocess

from rdopkg.cli import rdopkg

import test_common as common


RPMLINT_AVAILABLE = False
try:
    subprocess.call(["rpmlint", "--version"])
    RPMLINT_AVAILABLE = True
except Exception as e:
    pass


def _assert_sanity_out(o):
    assert 'ERRORS found:' in o
    assert 'E: buildarch-before-sources' in o


def _assert_rpmlint_out(o):
    # check expected errors were raised
    assert 'ERRORS found:' in o
    assert 'E: hardcoded-library-path' in o
    assert 'E: specfile-error warning:' in o
    assert 'prereq is deprecated: PreReq' in o
    # check expected warnings were raised
    assert 'WARNINGS found:' in o
    assert 'W: macro-in-%changelog' in o
    assert 'W: setup-not-in-prep' in o
    # make sure parsing rpmlint output was a success
    assert 'Failed to parse rpmlint output' not in o


def test_lint_sanity(tmpdir, capsys):
    dist_path = common.prep_spec_test(tmpdir, 'lint')
    rv = -1
    with dist_path.as_cwd():
        rv = rdopkg('lint', '--lint-checks', 'sanity')
    cap = capsys.readouterr()
    o = cap.out
    _assert_sanity_out(o)
    assert rv == 23


@pytest.mark.skipif('RPMLINT_AVAILABLE == False')
def test_lint_rpmlint(tmpdir, capsys):
    dist_path = common.prep_spec_test(tmpdir, 'lint')
    rv = -1
    with dist_path.as_cwd():
        rv = rdopkg('lint', '--lint-checks', 'rpmlint')
    cap = capsys.readouterr()
    o = cap.out
    _assert_rpmlint_out(o)
    assert rv == 23


@pytest.mark.skipif('RPMLINT_AVAILABLE == False')
def test_lint_all(tmpdir, capsys):
    # run `rdopkg lint` on tests/assets/spec/lint.spec
    dist_path = common.prep_spec_test(tmpdir, 'lint')
    rv = -1
    with dist_path.as_cwd():
        rv = rdopkg('lint')
    cap = capsys.readouterr()
    o = cap.out
    _assert_rpmlint_out(o)
    _assert_sanity_out(o)
    # linting problems are indicated by exit code 23
    assert rv == 23



#!/usr/bin/env python
"""
An example script that uses rdopkg.actionmods.rdoinfo to output a list of
currently maintained RDO projects.
"""

from distroinfo.query import filter_pkgs
from rdopkg.actionmods import rdoinfo


def list_projects():
    inforepo = rdoinfo.get_rdoinfo()
    info = inforepo.get_info()
    pkgs = info['packages']

    pkg_filter = {
        'name': '^openstack-'
    }
    pkgs = filter_pkgs(pkgs, pkg_filter)

    for pkg in pkgs:
        project = pkg.get('project') or pkg['name']
        print("### " + project)
        print("package: " + pkg['name'])
        print("code: " + pkg.get('upstream', 'N/A'))
        print("maintainers: " + " ".join(pkg['maintainers']))
        print("")


if __name__ == '__main__':
    list_projects()
